# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Internal utilities for testing, build, pipelines, and logging.
"""

from .testing import stream_handler  # noqa: F401
from .utils import experimental, b64_encode, b64_decode  # noqa: F401
from .eyesoff import is_eyesoff_helper, POLYMER_FEED  # noqa: F401



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import re
from pathlib import Path
from typing import List, Optional
from packaging.version import parse
from ruamel.yaml import YAML
import os
import multiprocessing

from shrike.build.core.command_line import Command

log = logging.getLogger(__name__)


class Register(Command):
    def __init__(self):
        super().__init__()

    def validate_branch(self) -> None:
        """
        Check whether the current source branch name matches the configured
        regular expression of branch name. Fail if it doesn't match.
        """
        log.info(f"Expected branch: {self.config.compliant_branch}")
        log.info(f"Current branch: {self.config.source_branch}")
        if re.match(self.config.compliant_branch, self.config.source_branch):
            log.info(f"Current branch matches configured regular expression.")
        else:
            raise ValueError(
                f"Current branch name doesn't match configured name pattern."
            )

    def find_signed_component_specification_files(self, dir=None) -> List[str]:
        """
        Find all signed component (AML format) generated in the `prepare` step. Return the
        absolute paths of these components in the format of a list of string.
        """
        if dir is None:
            dir = self.config.working_directory

        signed_component_spec_files = []
        # Find the path of spec files in the '.build' folders
        for spec_path in Path(dir).glob(
            "**/.build/" + Path(self.config.component_specification_glob).name
        ):
            # Check whether the component is signed by examining catalog files
            if (
                spec_path.parent.joinpath("catalog.json").exists()
                and spec_path.parent.joinpath("catalog.json.sig").exists()
            ):
                signed_component_spec_files.append(os.path.abspath(spec_path))
                log.info(f"Find a signed component for AML: {spec_path}")
            elif spec_path.parent.joinpath(".build.cat").exists():
                log.info(f"Find a signed component for Aether: {spec_path}")
            else:
                log.warning(f"Find an unsigned component: {spec_path}")
            log.info(str(spec_path.parent.joinpath("catalog.json")))

        if len(signed_component_spec_files) == 0:
            log.info("Cannot find any signed components for AML.")
        else:
            log.info(
                f"Find {len(signed_component_spec_files)} signed components for AML."
            )

        return signed_component_spec_files

    def list_registered_component(self) -> None:
        """
        Log all registered component in the attached workspace by using az ml command.
        """
        list_registered_component_success = self.execute_azure_cli_command(
            f"ml component list -o table"
        )
        if not list_registered_component_success:
            self.register_error(f"Error when listing registered components.")

    def register_all_signed_components(self, files: List[str], registry: Optional[str] = None) -> None:
        """
        Run register_component_with_status() in parallel with nb_cores threads
        """
        if files:
            nb_cores = self.nb_cores

            if nb_cores == 1:
                for component in files:
                    self.register_component_with_status(component, registry)
            else:
                log.info(f"Batch register with {nb_cores} threads.")
                files_with_registry = [(f, registry) for f in files]
                pool = multiprocessing.Pool(processes=nb_cores)
                pool.starmap(self.register_component_with_status, files_with_registry)
                pool.close()
                pool.join()

    def register_component_with_status(self, component, registry: Optional[str] = None) -> None:
        """
        For one of the signed component specification file, run `az ml component create`,
        and register the status (+ register error if registration failed).
        """
        register_command = self.register_component_command(component, registry)

        register_component_success = self.execute_azure_cli_command(
            command=register_command,
            stderr_is_failure=False,
            fail_if_version_exists=self.config.fail_if_version_exists,
        )
        if register_component_success:
            log.info(f"Component {component} is registered.")
            self.register_component_status(component, "register", "succeeded")
        else:
            self.register_component_status(component, "register", "failed")
            self.register_error(f"Error when registering component {component}.")

    def register_component_command(self, component, registry: Optional[str] = None):
        register_command = f"ml component create --file {component}"
        set_default_version = False
        component_raw_version = self.read_component_version(component)

        if self.config.all_component_version:
            register_command += f" --version {self.config.all_component_version}"
            component_raw_version = self.config.all_component_version
            log.info(
                f"Overwrite the component version with the specified value {self.config.all_component_version}"
            )

        if registry:
            register_command += f" --registry {registry}"

        try:
            component_version = parse(component_raw_version)
            set_default_version = (
                component_version.base_version == component_raw_version
            )
        except:
            log.error(f"{component_raw_version} is not a valid version number.")
        if set_default_version:
            log.info(
                f"Component {component} version {component_raw_version} is production-ready. Setting as default."
            )
            register_command += f" --label default"
        else:
            log.info(
                f"Component {component} version {component_raw_version} is not production-ready. NOT setting as default."
            )

        log.info(f"Register command is {register_command}")
        return register_command

    def read_component_version(self, yaml_file: str) -> str:
        yaml = YAML(typ="safe")
        with open(yaml_file, "r") as file:
            spec = yaml.load(file)
        try:
            version = spec["version"]
            log.info(f"Component {yaml_file} has version {version}.")
            return version
        except KeyError:
            log.warning(
                "Component does not have version attribute, attempting to read module version."
            )
            return spec["amlModuleIdentifier"]["moduleVersion"]

    def run_with_config(self):
        """
        Running component registration logic.
        """
        self.telemetry_logging(command="register")

        self.validate_branch()
        self.ensure_component_cli_installed()

        component_path = self.find_signed_component_specification_files()
        if len(component_path) > 0:
            for workspace_id in self.config.workspaces:
                log.info(f"Start registering signed components in {workspace_id}")
                self.attach_workspace(workspace_id)

                log.info("List of components in workspace before current registration.")
                self.list_registered_component()

                self.register_all_signed_components(files=component_path)

                log.info("List of components in workspace after current registration.")
                self.list_registered_component()
            for registry in self.config.registries:
                log.info(f"Start registering signed components in {registry}")
                self.register_all_signed_components(files=component_path, registry=registry)


if __name__ == "__main__":
    Register().run()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Shared utils.
"""
import logging
import base64
from functools import wraps

log = logging.getLogger(__name__)


def experimental(
    message="This is an experimental feature and could change at any time.",
):
    def wrapper(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            log.warning(message)
            return func(*args, **kwargs)

        return new_func

    return wrapper


def b64_encode(s: str, encoding: str = "ascii") -> str:
    """
    Produce base-64 encoded version of `s`.
    """
    s_bytes = s.encode(encoding)
    b64_bytes = base64.b64encode(s_bytes)
    s_encoded = b64_bytes.decode(encoding)
    return s_encoded


def b64_decode(s: str, encoding: str = "ascii") -> str:
    """
    Return str corresponding to base-64 decoded, YAML
    deserialized, version of `s`.
    """
    b64_bytes = s.encode(encoding)
    s_bytes = base64.b64decode(b64_bytes)
    s_decoded = s_bytes.decode(encoding)
    return s_decoded



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Helper code to run Dask distributed scripts [EXPERIMENTAL]
"""
import asyncio
import os
import socket
import sys
from typing import Optional, cast

import psutil
from shrike._core import experimental
from shrike.distributed import EXPERIMENTAL_WARNING_MSG
from shrike.compliant_logging import DataCategory
from shrike.distributed.cluster_auto_setup import ClusterAutoSetupHandler
from dask.distributed import Client, Scheduler, Worker, performance_report


# pragma: no cover
class DaskClusterSetupHandler(ClusterAutoSetupHandler):
    def __init__(
        self,
        process_count_per_node: Optional[int] = None,
        thread_count_per_process: Optional[int] = None,
        dask_report_link: Optional[str] = None,
    ):
        """Constructor for Dask Cluster

        Args:
            process_count_per_node (int): Number of processes available per node.
                Defaults to `None`.
                If `None` the machine cpu count defines the number of processes.
            thread_count_per_process (int): Number of threads available per process.
                Defaults to `None`. If `None` the number of threads is defined by the
                division of cpu count by number of processes.
            dask_report_link (str): Path where dask dashboard report will be written to.
                Defaults to None. When None, dask dashboard is not reported/written to
                a file.

        Example:

            # Starts a Cluster using as many processes as cpu count
            # (Ideal to avoid workloads suffering from GIL lock):
            .. code-block:: python

            from shrike.distributed.dask import DaskClusterSetupHandler

            with DaskClusterSetupHandler() as azure_cluster:
                client = azure_cluster.client
                logger.info(f"Dask {client}")


            # Starts a Cluster using 1 process per machine,
            # and give to it as many threads as cpu count:
            .. code-block:: python

            from shrike.distributed.dask import DaskClusterSetupHandler

            with DaskClusterSetupHandler(process_count_per_node=1) as azure_cluster:
                client = azure_cluster.client
                logger.info(f"Dask {client}")


        """
        super().__init__()
        self.head_port = 45462
        self.node_count = int(cast(str, os.getenv("OMPI_COMM_WORLD_SIZE")))
        self.rank = int(cast(str, os.getenv("OMPI_COMM_WORLD_RANK")))
        self.ip = socket.gethostbyname(socket.gethostname())

        self.head_address = None
        if process_count_per_node is None:
            self.process_count_per_node = psutil.cpu_count()
        else:
            self.process_count_per_node = process_count_per_node

        # How many threads is available per worker
        if thread_count_per_process is None:
            self.thread_count_per_process = int(
                psutil.cpu_count() / self.process_count_per_node
            )
        else:
            self.thread_count_per_process = int(thread_count_per_process)
        self.worker_ram = int(
            0.90 * psutil.virtual_memory().available / self.process_count_per_node
        )
        self.worker_count = (self.node_count - 1) * self.process_count_per_node

        # output dashboard
        if dask_report_link is not None:
            self.dask_performance_report_file = os.path.join(  # type: ignore
                dask_report_link, "dask-report.html"
            )
        else:
            self.dask_performance_report_file = None

    #################
    # SETUP METHODS #
    #################

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def setup_head_node(self):
        """Setup to run only on head node"""
        self.logger.info(
            f"{self.__class__.__name__}.setup_head_node() called to set up HEAD node."
        )
        # create setup config
        self.head_address = socket.gethostbyname(socket.gethostname())

        # record what's needed to setup cluster nodes
        self.setup_config_add_key("head_address", self.head_address)

        self.pid = os.fork()  # type: ignore
        if self.pid > 0:
            self.rank = "0.1"

            async def run_scheduler():
                async with Scheduler(
                    port=self.head_port, scheduler_file="scheduler.json"
                ) as scheduler:
                    await scheduler.finished()

            asyncio.get_event_loop().run_until_complete(run_scheduler())
            self.logger.info(
                f"[R{self.rank}]SCHEDULER IS OFFLINE", category=DataCategory.PUBLIC
            )
            self.multinode_driver.finalize()
            sys.exit(0)

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def setup_cluster_node(self):
        """Setup to run only on non-head cluster nodes"""
        self.logger.info(f"{self.__class__.__name__}.setup_cluster_node() called")
        self.head_address = self.setup_config_get_key("head_address")

    def __enter__(self):
        self.initialize_run()

        for line in self.__repr__().split("\n"):
            self.logger.info(line, category=DataCategory.PUBLIC)

        if self.rank != 0:
            self.logger.info("Running worker...", category=DataCategory.PUBLIC)

            async def run_worker():
                async with Worker(
                    scheduler_ip=self.head_address,
                    scheduler_port=self.head_port,
                    nthreads=self.thread_count_per_process,
                    memory_limit=self.worker_ram,
                    local_directory=os.getenv("AZ_BATCHAI_JOB_TEMP"),
                    name=f"{self.rank}",
                ) as worker:
                    self.logger.info(
                        f"[R{self.rank}] - {worker}", category=DataCategory.PUBLIC
                    )
                    await worker.finished()

            if self.process_count_per_node > 1:
                for i in range(self.process_count_per_node - 1):
                    pid = os.fork()  # type: ignore
                    if pid > 0:
                        self.rank = f"{self.rank}_{i}"
                        asyncio.get_event_loop().run_until_complete(run_worker())
                        self.logger.info(
                            f"[R{self.rank}]Worker IS OFFLINE",
                            category=DataCategory.PUBLIC,
                        )
                        self.multinode_driver.finalize()
                        sys.exit(0)

            # Shut down when down
            asyncio.get_event_loop().run_until_complete(run_worker())
            self.logger.info(
                f"[R{self.rank}]Worker IS OFFLINE", category=DataCategory.PUBLIC
            )
            self.multinode_driver.finalize()
            sys.exit(0)
        else:
            self.logger.info("Starting Dask Client...", category=DataCategory.PUBLIC)
            self.client = Client(f"tcp://{self.head_address}:{self.head_port}")
            self.logger.info(
                f"Waiting for {self.worker_count} workers up to 10s",
                category=DataCategory.PUBLIC,
            )
            self.client.wait_for_workers(self.worker_count, 10)

            if self.dask_performance_report_file is not None:
                self.dask_performance_report = performance_report(
                    filename=self.dask_performance_report_file
                )
                self.dask_performance_report.__enter__()

        return self

    def __repr__(self) -> str:
        string = (
            "Cluster info\n"
            f"\t├─>\t\t rank = {self.rank} \n"
            f"\t├─>\t\t ip = {self.ip} \n"
            f"\t├─>\t\t scheduler_ip = {self.head_address} \n"
            f"\t├─>\t\t node_count = {self.node_count} \n"
            f"\t├─>\t\t process_count_per_node = {self.process_count_per_node} \n"
            f"\t├─>\t\t thread_count_per_process = {self.thread_count_per_process} \n"
            f"\t├─>\t\t worker_ram = {(self.worker_ram/(2**30)):.2f} GB \n"
            f"\t├─>\t\t worker_count = {self.worker_count} \n"
        )
        return string

    def __exit__(self, type, value, traceback):

        self.logger.info(
            f"Exiting context from rank {self.rank}", category=DataCategory.PUBLIC
        )
        if self.rank == 0:

            self.logger.info(
                f"Dask dashboard saved to {self.dask_performance_report_file}",
                category=DataCategory.PUBLIC,
            )
            self.logger.info("Shutting dask down...", category=DataCategory.PUBLIC)
            self.client.shutdown()  # This will also close the workers

            if os.path.exists("scheduler.json"):
                os.remove("scheduler.json")
            # self.multinode_driver.finalize()
            if self.dask_performance_report_file is not None:
                self.dask_performance_report.__exit__(type, value, traceback)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Constant values used by this library.
"""

from enum import Enum


class DataCategory(Enum):
    """
    Enumeration of data categories in compliant machine learning.

    Values:
    - PRIVATE: data which is private. Researchers may not view this.
    - PUBLIC: data which may safely be viewed by researchers.
    """

    PRIVATE = 1
    PUBLIC = 2



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import platform

from shrike.compliant_logging.constants import DataCategory


def provide_system_info(
    logger: logging.Logger = None,
    library_checks: list = None,
) -> None:
    """
    Provides a list of information about the current system.

    Args:
        logger (logging.Logger, optional): Logger to use. Defaults to None.
        additional_libs (list, optional): List of libraries that should be checked
            for import. Defaults to None.
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    # retrieve information about the current system
    uname = platform.uname()
    python_version = platform.python_version()
    config_str = "### System Config ###"
    logger.info(config_str, category=DataCategory.PUBLIC)
    logger.info(f"| CPU     = {uname.processor}", category=DataCategory.PUBLIC)
    logger.info(f"| Machine = {uname.machine}", category=DataCategory.PUBLIC)
    logger.info(
        f"| System  = {uname.system} ({uname.release})",
        category=DataCategory.PUBLIC,
    )
    logger.info(f"| Python  = {python_version}", category=DataCategory.PUBLIC)
    logger.info("#" * len(config_str), category=DataCategory.PUBLIC)

    # check for libraries
    if library_checks is not None:
        # iterate through all libs
        for lib in library_checks:
            try:
                __import__(lib)
                logger.info(f"Library {lib} available", category=DataCategory.PUBLIC)
            except ModuleNotFoundError:
                logger.warning(
                    f"Library {lib} is not found and could not be imported",
                    category=DataCategory.PUBLIC,
                )
            except Exception as ex:
                logger.warning(
                    f"Library {lib} could not be imported: {ex}",
                    category=DataCategory.PUBLIC,
                )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Python utilities that help users to manage, validate
and submit AML pipelines
"""


from .pipeline_helper import AMLPipelineHelper  # noqa: F401
from .federated_learning import FederatedPipelineBase, StepOutput  # noqa: F401
from .ray_actor import ray_actor_on_shrike, b64_decode, b64_encode  # noqa: F401
from .argparser_utils import get_args_from_component_spec  # noqa: F401



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
"Driver" for MPI communication
"""
import logging
import traceback
from dataclasses import dataclass


@dataclass
class MultiNodeConfig:
    world_size: int = 1
    world_rank: int = 0
    multinode_available: bool = False
    main_node: bool = True


class MultiNodeMPIDriver:
    """Handling MPI initialization in a separate class
    so we can patch/mock it during unit testing of MultiNodeScript"""

    def __init__(self, mpi_init_mode=None):
        """Constructor"""
        self.logger = logging.getLogger(__name__)
        self._multinode_config = None
        self._mpi_module = None
        self._mpi_init_mode = mpi_init_mode

    def _mpi_import(self):
        self.logger.info("Importing mpi4py...")
        # doing our own initialization of MPI to have fine-grain control
        import mpi4py

        mpi4py.rc.initialize = False
        mpi4py.rc.finalize = False
        from mpi4py import MPI

        return MPI

    def get_multinode_config(self) -> MultiNodeConfig:
        """Get internal multinode config"""
        if self._multinode_config:
            return self._multinode_config
        else:
            raise Exception("Multinode config is None, use initialize() first.")

    def get_comm(self):
        """Returns the communicator"""
        if self._mpi_module:
            return self._mpi_module.COMM_WORLD
        else:
            raise Exception(
                "MPI.COMM_WORLD was requested but MPI was never initialized"
                + "please use mpi_init_mode!=None."
            )

    def initialize(self):
        """Initialize the driver"""
        self.logger.info(f"Call to {self.__class__.__name__}.initialize()")
        self._mpi_module = self._mpi_import()

        # init mpi and use comm to detect mpi config
        try:
            if self._mpi_init_mode is None:
                self.logger.info("Running MPI.Init()")
                self._mpi_module.Init()
            else:
                self.logger.info(
                    f"Running MPI.Init_thread(required={self._mpi_init_mode})"
                )
                self._mpi_module.Init_thread(required=self._mpi_init_mode)
        except self._mpi_module.Exception:
            self.logger.warning(
                f"Exception occured during MPI init:\n{traceback.format_exc()}"
            )

        _comm = self._mpi_module.COMM_WORLD
        try:
            self._multinode_config = MultiNodeConfig(
                _comm.Get_size(),  # world_size
                _comm.Get_rank(),  # world_rank
                (_comm.Get_size() > 1),  # mpi_available
                (_comm.Get_rank() == 0),  # main_node
            )
            self.logger.info(f"MPI detection results: {self._multinode_config}")
        except BaseException:
            self.logger.warning(traceback.format_exc())
            self._multinode_config = MultiNodeConfig(
                1,  # world_size
                0,  # world_rank
                False,  # mpi_available
                True,  # main_node
            )
            self.logger.critical(
                "MPI detection failed, switching to single node:"
                + f" {self._multinode_config}"
                + f" see traceback below:\n{traceback.format_exc()}"
            )

    def finalize(self):
        """Finalize/close resources used by the driver"""
        self.logger.info(f"Call to {self.__class__.__name__}.finalize()")
        if self._mpi_module is None:
            self.logger.warning(
                "MPIHandler.finalize() was called but MPI was never instanciated"
            )
        elif self._mpi_module.Is_initialized() and not self._mpi_module.Is_finalized():
            self.logger.info("MPI was initialized, calling MPI.finalize()")
            self._mpi_module.Finalize()
        else:
            self.logger.warning(
                "MPIHandler.finalize() was called,"
                + f" but MPI.Is_initialized={self._mpi_module.Is_initialized()}"
                + f"and MPI.Is_finalized={self._mpi_module.Is_finalized()}"
            )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""
This script is modified based on [aml_adapter.py](https://eemo.visualstudio.com/TEE/_git/TEEGit?path=/Offline/GIS/src/libraries/gis-common-argument/gis/common/argument/aml_adapter.py)
"""
import logging
import re
from argparse import ArgumentParser, Namespace
from typing import Dict, Union
from ruamel.yaml import YAML

from shrike.compliant_logging.exceptions import (
    PublicArgumentTypeError,
    PublicTypeError,
    PublicRuntimeError,
)
from shrike.compliant_logging import DataCategory


def str_to_bool(val: Union[bool, str]) -> bool:
    """
    Resolving boolean arguments if they are not given in the standard format

    Arguments:
        val (bool or string): boolean argument type
    Output:
        bool: the desired value {True, False}

    """
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        if val.lower() in ("yes", "true", "t", "y", "1"):
            return True
        elif val.lower() in ("no", "false", "f", "n", "0"):
            return False
    raise PublicArgumentTypeError("Boolean value expected.")


TYPE_MAPPING = {
    "String": str,
    "string": str,
    "AnyDirectory": str,
    "Integer": int,
    "integer": int,
    "int": int,
    "Boolean": str_to_bool,
    "boolean": str_to_bool,
    "Bool": str_to_bool,
    "bool": str_to_bool,
    "Float": float,
    "float": float,
    "Enum": str,
    "enum": str,
    "path": str,
}


def get_args_from_component_spec(
    component_spec_file_name: str, logger: logging.Logger
) -> Namespace:
    """create and apply the ArgumentParser based on component_spec_file_name

    Parameters
    ----------
    component_spec_file_name: the file name of component spec

    Returns
    -------
    args parsed from command
    """
    try:
        logger.info(
            "======== Parsing input arguments ========", category=DataCategory.PUBLIC
        )
    except TypeError:
        # TypeError: _log() got an unexpected keyword argument 'category'
        raise PublicTypeError(
            "The auto arg parser API only works with [category-aware logging](https://shrike-docs.com/compliant_logging/). Please call `enable_compliant_logging()` before creating a logger."
        )

    with open(component_spec_file_name, mode="r", encoding="utf-8") as stream:
        spec = YAML(typ="safe").load(stream)
        command_str = ""
        if "command" in spec:
            command_str = spec["command"]
        elif "hdinsight" in spec:
            command_str = spec["hdinsight"]["args"]
        elif "launcher" in spec:
            command_str = spec["launcher"]["additional_arguments"]
        elif "parallel" in spec:
            command_str = spec["parallel"]["args"]

        logger.info(command_str, category=DataCategory.PUBLIC)
        parser = _create_argument_parser(
            command_str, spec["inputs"], spec.get("outputs", None)
        )
    args, _ = parser.parse_known_args()
    _log_args(args, logger)

    return args


def _log_args(args: Namespace, logger: logging.Logger = None) -> None:
    """log the arguments for the debugging purpose"""
    logger.info(
        "======== Arguments from command ========", category=DataCategory.PUBLIC
    )
    for name, value in vars(args).items():
        logger.info("name=%s, value=%s" % (name, value), category=DataCategory.PUBLIC)


def _create_argument_parser(
    command: str, inputs: Dict[str, Dict[str, str]], outputs: Dict[str, Dict[str, str]]
) -> ArgumentParser:
    """create the argument parser from component spec: command, inputs and outputs"""
    parser = ArgumentParser()
    command_dict = _parse_command(command)

    for arg_name, arg_setting in command_dict.items():
        inputs_or_outputs, field = arg_setting.strip().split(".")
        if inputs_or_outputs == "inputs":
            _add_argument(parser, arg_name, inputs[field])
        elif inputs_or_outputs == "outputs":
            _add_argument(parser, arg_name, outputs[field])
        else:
            raise PublicRuntimeError("unknown fields in spec command")

    return parser


def _parse_command(command: str) -> Dict[str, str]:
    """parse the command to dict: key is the optional argument,
    value is either inputs or outputs

    Examples
    --------
    command = 'python3 run.py --input {inputs.input_path} --input_file {inputs.input_file} --output {outputs.output_file}
    [--apply {inputs.apply}] [--select {inputs.select}]'

    return = {'--input': 'inputs.input_path', '--input_file': 'inputs.input_file', '--output': 'outputs.output_file',
              '--apply': 'inputs.apply', '--select': 'inputs.select'}
    """
    pattern = r"(--[a-zA-Z0-9_]+)\s+\{([a-zA-Z0-9\._]+)\}"
    matches = re.findall(pattern, command)
    return dict(matches)


def _add_argument(
    parser: ArgumentParser, argument_name: str, setting: Dict[str, str]
) -> None:
    """mapping from AML component spec keywords to `add_argument` keywords:

    description -> help
    optional -> required
    type -> type
    default -> default
    enum -> choices
    """
    help_info = setting.get("description", "no help info for this argument")
    setting_type = setting.get("type", "String")
    type_ = TYPE_MAPPING.get(setting_type, str)
    required = False if setting.get("optional", False) else True
    default_value = setting.get("default", None)
    choices = setting.get("enum", None)

    parser.add_argument(
        argument_name,
        required=required,
        type=type_,
        help=help_info,
        default=default_value,
        choices=choices,
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""


from azure.ml.component import Component
from azure.ml.component._core._component_definition import ComponentDefinition
from dataclasses import dataclass, field
import os
import logging
from typing import Optional, List, Tuple
from packaging.specifiers import SpecifierSet
from packaging.version import Version, InvalidVersion

from shrike.pipeline.aml_connect import current_workspace


log = logging.getLogger(__name__)


@dataclass
class module_reference:
    key: Optional[
        str
    ] = None  # use as internal key to reference module (if None, use name)
    name: Optional[str] = None  # None if module exists only locally?
    source: Optional[str] = None  # This config is deprecated. Please do not use.
    yaml: Optional[str] = None
    version: Optional[str] = None
    from_workspace: Optional[
        bool
    ] = False  # If set to True and it's a remote component, shrike will only attempt to load it from workspace (not from registry or feed)


@dataclass
class module_manifest:
    manifest: List[module_reference] = field(default_factory=list)
    feeds: List[str] = field(default_factory=list)
    registries: List[str] = field(default_factory=list)


@dataclass
class module_loader_config:  # pylint: disable=invalid-name
    """Config for the AMLModuleLoader class"""

    use_local: Optional[str] = None
    use_local_except_for: Optional[bool] = None
    force_default_module_version: Optional[str] = None
    force_all_module_version: Optional[str] = None
    local_steps_folder: str = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "tests",
            "tests_pipeline",
            "sample",
            "steps",
        )
    )
    use_remote_when_component_not_in_manifest: Optional[bool] = False
    loading_order: Optional[str] = "registry, feed, workspace"


class AMLModuleLoader:
    """Helper class to load modules from within an AMLPipelineHelper."""

    def __init__(self, config):
        """Creates module instances for AMLPipelineHelper.

        Args:
            config (DictConfig): configuration options
        """
        self.use_local_except_for = (
            config.module_loader.use_local_except_for
            if "use_local_except_for" in config.module_loader
            else None
        )
        if "use_local" not in config.module_loader:
            self.use_local = []
        elif config.module_loader.use_local is None:
            self.use_local = []
        elif config.module_loader.use_local == "*":
            self.use_local = "*"
        elif isinstance(config.module_loader.use_local, str):
            self.use_local = [
                x.strip() for x in config.module_loader.use_local.split(",")
            ]
            if not _check_use_local_syntax_valid(self.use_local):
                raise ValueError(
                    f'Invalid value for `use_local`. Please follow one of the four patterns: \n1) use_local="", all modules are remote\n2) use_local="*", all modules are local\n3) use_local="MODULE_KEY_1, MODULE_KEY_2", only MODULE_KEY_1, MODULE_KEY_2 are local, everything else is remote\n4) use_local="!MODULE_KEY_1, !MODULE_KEY_2", all except for MODULE_KEY_1, MODULE_KEY_2 are local'
                )
            self.use_local_except_for = self.use_local[0].startswith("!")

        self.force_default_module_version = (
            config.module_loader.force_default_module_version
            if "force_default_module_version" in config.module_loader
            else None
        )
        self.force_all_module_version = (
            config.module_loader.force_all_module_version
            if "force_all_module_version" in config.module_loader
            else None
        )
        self.local_steps_folder = config.module_loader.local_steps_folder
        self.use_remote_when_component_not_in_manifest = (
            config.module_loader.use_remote_when_component_not_in_manifest
            if "use_remote_when_component_not_in_manifest" in config.module_loader
            else None
        )
        self.module_cache = {}

        self.feeds: List[str] = []
        self.registries: List[str] = []

        self.loading_order = (
            [x.strip().lower() for x in config.module_loader.loading_order.split(",")]
            if "loading_order" in config.module_loader
            else ["registry", "feed", "workspace"]
        )

        # internal manifest built from yaml config
        self.modules_manifest = {}
        self.load_config_manifest(config)

        initialization_info_string = (
            "AMLModuleLoader initialized ("
            f"use_local={self.use_local}"
            f", force_default_module_version={self.force_default_module_version}"
            f", force_all_module_version={self.force_all_module_version}"
            f", local_steps_folder={self.local_steps_folder}"
            f", use_remote_when_component_not_in_manifest={self.use_remote_when_component_not_in_manifest}"
            f", manifest={list(self.modules_manifest.keys())}"
            f", loading_order={self.loading_order}"
            ")"
        )
        log.info(initialization_info_string)

    def load_config_manifest(self, config):
        """Fills the internal module manifest based on config object"""
        for entry in config.modules.manifest:
            if entry.key:
                module_key = entry.key
            elif entry.name:
                module_key = entry.name
            else:
                raise Exception(
                    "In module manifest, you have to provide at least key or name."
                )

            self.modules_manifest[module_key] = entry

        for feed in config.modules.get("feeds", []):
            self.feeds.append(feed)

        for registry in config.modules.get("registries", []):
            self.registries.append(registry)

    def is_local(self, module_name):
        """Tests is module is in local list"""
        if self.use_local == "*":
            return True
        if self.use_local_except_for:
            return "!" + module_name not in self.use_local
        else:
            return module_name in self.use_local

    def module_in_cache(self, module_cache_key):
        """Tests if module in internal cache (dict)"""
        return module_cache_key in self.module_cache

    def get_from_cache(self, module_cache_key):
        """Gets module class from internal cache (dict)"""
        log.debug(f"Using cached module {module_cache_key}")
        return self.module_cache.get(module_cache_key, None)

    def put_in_cache(self, module_cache_key, module_class):
        """Puts module class in internal cache (dict)"""
        self.module_cache[module_cache_key] = module_class

    def verify_manifest(self, modules_manifest):
        """Tests a module manifest schema"""
        errors = []

        for (k, module_entry) in modules_manifest.items():
            # TODO: merge error checking code with processing code so we do all this in one pass
            if self.is_local(k):
                if "yaml_spec" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a yaml_spec for your module to use_local=['{k}']"
                    )
                elif not os.path.isfile(
                    module_entry["yaml_spec"]
                ) and not os.path.isfile(
                    os.path.join(self.local_steps_folder, module_entry["yaml_spec"])
                ):
                    errors.append(
                        "{}: Could not find yaml spec {} for use_local=['{}']".format(
                            k, module_entry["yaml_spec"], k
                        )
                    )
            else:
                if "remote_module_name" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a name for your module to use_local=False"
                    )
                if "namespace" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a namespace for your module to use_local=False"
                    )
                if ("version" not in module_entry) and (
                    self.force_default_module_version or self.force_all_module_version
                ):
                    errors.append(
                        f"{k}: You need to specify a version for your module to use_local=False, or use either force_default_module_version or force_all_module_version in config"
                    )

        return errors

    def load_local_module(self, module_spec_path):
        """Creates one module instance.

        Args:
            module_spec_path (str): path to local module yaml spec

        Returns:
            object: module class loaded
        """
        module_cache_key = module_spec_path
        if self.module_in_cache(module_cache_key):
            return self.get_from_cache(module_cache_key)

        log.info("Building module from local code at {}".format(module_spec_path))
        if not os.path.isfile(module_spec_path):
            module_spec_path = os.path.join(self.local_steps_folder, module_spec_path)
        loaded_module_class = Component.from_yaml(yaml_file=module_spec_path)
        self.put_in_cache(module_cache_key, loaded_module_class)

        return loaded_module_class

    def load_module_from_registry(self, module_name, module_version, module_cache_key):
        log.info(f"Attempting to load {module_name} from registries...")
        for registry in self.registries:
            try:
                log.info(f"Loading component {module_name} from registry {registry}")

                module_version = solve_module_version(
                    module_name, module_version, registry=registry
                )

                loaded_module_class = Component.load(
                    current_workspace(),
                    name=module_name,
                    version=module_version,
                    registry=registry,
                )
                self.put_in_cache(module_cache_key, loaded_module_class)
                return loaded_module_class

            except BaseException as e:
                log.debug(f"Exception: {e}")

    def load_module_from_feed(self, module_name, module_version, module_cache_key):
        log.info(f"Attempting to load {module_name} from feeds...")
        for feed in self.feeds:
            try:
                log.info(f"Loading component {module_name} from feed {feed}")

                module_name = f"azureml.feed://{feed}/{module_name}"

                module_version = solve_module_version(
                    module_name, module_version, workspace=current_workspace()
                )

                loaded_module_class = Component.load(
                    current_workspace(),
                    name=module_name,
                    version=module_version,
                )
                self.put_in_cache(module_cache_key, loaded_module_class)
                return loaded_module_class

            except BaseException as e:
                log.debug(f"Exception: {e}")

    def load_module_from_workspace(
        self, module_name, module_version, module_cache_key, module_namespace
    ):
        log.info(f"Attempting to load {module_name} from workspace...")
        try:
            # try without namespace first
            module_version = solve_module_version(
                module_name, module_version, workspace=current_workspace()
            )

            loaded_module_class = Component.load(
                current_workspace(),
                name=module_name,
                version=module_version,
            )
        except BaseException as e:
            # save the exception to raise it if namespace not provided
            if not module_namespace:
                raise e
        if module_namespace:
            log.info(
                f"    Trying to load module {module_name} with namespace {module_namespace}."
            )
            module_name = module_namespace + "://" + module_name
            module_version = solve_module_version(
                module_name, module_version, workspace=current_workspace()
            )

            loaded_module_class = Component.load(
                current_workspace(),
                name=module_name,
                version=module_version,
            )

        self.put_in_cache(module_cache_key, loaded_module_class)
        return loaded_module_class

    def load_prod_module(
        self, module_name, module_version, from_workspace=False, module_namespace=None
    ):
        """Creates one module instance.

        Args:
            module_name (str) : module name
            module_version (str) : module version
            from_workspace (bool) : attempt to load from workspace only

        Returns:
            object: module class loaded
        """
        if self.force_all_module_version:
            module_version = self.force_all_module_version
        else:
            module_version = module_version or self.force_default_module_version

        module_cache_key = f"{module_name}:{module_version}"
        if self.module_in_cache(module_cache_key):
            return self.get_from_cache(module_cache_key)

        log.info(
            f"Loading remote module {module_cache_key} (name={module_name}, version={module_version}, namespace={module_namespace})"
        )

        if from_workspace:
            return self.load_module_from_workspace(
                module_name, module_version, module_cache_key, module_namespace
            )
        for i in self.loading_order:
            if i == "registry":
                loaded_module_class = self.load_module_from_registry(
                    module_name, module_version, module_cache_key
                )
            elif i == "feed":
                loaded_module_class = self.load_module_from_feed(
                    module_name, module_version, module_cache_key
                )
            elif i == "workspace":
                loaded_module_class = self.load_module_from_workspace(
                    module_name, module_version, module_cache_key, module_namespace
                )
            else:
                raise ValueError(
                    f"Loading order contains incorrect key {i}. Please use a comma-separated str to order 'registry', 'feed', and 'workspace'."
                )
            if loaded_module_class:
                return loaded_module_class

    def get_module_manifest_entry(
        self, module_key, modules_manifest=None
    ) -> Tuple[dict, Optional[str], bool]:
        """Gets a particular entry in the module manifest.

        Args:
            module_key (str): module key from the manifest
            modules_manifest (dict): manifest from required_modules() [DEPRECATED]

        Returns:
            module_entry (dict): module manifest entry (if no entry for this module key in the manifest, only the module key is returned)
            module_namespace (str | None): module namespace for legacy modules
            is_in_manifest (bool): true if the module key can be found in the manifest
        """
        if module_key in self.modules_manifest:
            module_entry = self.modules_manifest[module_key]
            module_namespace = None
            is_in_manifest = True
        elif modules_manifest and module_key in modules_manifest:
            log.warning(
                f"We highly recommend substituting the `required_modules` method by the modules.manifest configuration."
            )
            module_entry = modules_manifest[module_key]
            # map to new format
            module_entry["yaml"] = module_entry["yaml_spec"]
            module_entry["name"] = module_entry["remote_module_name"]
            module_namespace = module_entry.get("namespace", None)
            is_in_manifest = True
        else:
            module_entry = {}
            module_entry["name"] = module_key
            module_namespace = None
            is_in_manifest = False
            if not (self.use_remote_when_component_not_in_manifest):
                log.warning(
                    f"Module key '{module_key}' could not be found in modules.manifest configuration or in required_modules() method. If you want to try and load it from the workspace, set 'module_loader.use_remote_when_component_not_in_manifest' to True in the config."
                )

        return module_entry, module_namespace, is_in_manifest

    def load_module(self, module_key, modules_manifest=None):
        """Loads a particular module from the manifest.

        Args:
            module_key (str): module key from the manifest
            modules_manifest (dict): manifest from required_modules() [DEPRECATED]

        Returns:
            object: module class loaded
        """
        module_entry, module_namespace, is_in_manifest = self.get_module_manifest_entry(
            module_key, modules_manifest
        )
        if is_in_manifest:
            if self.is_local(module_key):
                loaded_module = self.load_local_module(module_entry["yaml"])
            else:
                loaded_module = self.load_prod_module(
                    module_entry["name"],
                    module_entry["version"],
                    from_workspace=module_entry["from_workspace"]
                    if "from_workspace" in module_entry
                    else False,
                    module_namespace=module_namespace,
                )
        else:
            log.warning(
                f"The component '{module_key}' cannot be found in the manifest. Attempting to load the remote copy."
            )
            loaded_module = self.load_prod_module(module_entry["name"], None)
        return loaded_module

    def load_modules_manifest(self, modules_manifest):
        """Creates module instances from modules_manifest.

        Args:
            modules_manifest (dict): manifest of modules to load

        Returns:
            dict: modules loaded, keys are taken from module_manifest.

        Raises:
            Exception: if loading module has an error or manifest is wrong.
        """
        log.info(f"Loading module manifest (use_local={self.use_local})")
        test_results = self.verify_manifest(modules_manifest)
        if test_results:
            raise Exception(
                "Loading modules from manifest raised errors:\n\nMANIFEST: {}\n\nERRORS: {}".format(
                    modules_manifest, "\n".join(test_results)
                )
            )

        loaded_modules = {}
        for module_key in modules_manifest:
            log.info(f"Loading module {module_key} from manifest")
            loaded_modules[module_key] = self.load_module(module_key, modules_manifest)

        return loaded_modules


def solve_module_version(module_name, module_version, workspace=None, registry=None):

    if module_version is None:
        return module_version

    try:
        module_version_PEP440 = Version(module_version)
        if str(module_version_PEP440) != module_version:
            log.warning(
                "We suggest adopting PEP440 versioning for your component {module_name}!"
            )

    except InvalidVersion as e:
        log.info(f"{module_version} is a version constraint. Try to solve it...")

        spec = SpecifierSet(module_version)
        if registry:
            components = ComponentDefinition.list(
                registry_name=registry, name=module_name
            )
        elif workspace:
            components = ComponentDefinition.list(workspace=workspace, name=module_name)

        versions = []
        for component in components:
            version = component.version
            version_PEP440 = Version(version)
            if str(version_PEP440) != version:
                log.warning(
                    f"Version {version} does not follow PEP440 versioning, skipping ..."
                )
            else:
                versions.append(version_PEP440)

        compatible_versions = list(spec.filter(versions))
        if compatible_versions:
            module_version = max(compatible_versions)
            log.info(f"Solved version for {module_name} is {module_version}.")

        else:
            raise ValueError(
                f"No version exists for the constraint {module_version}. Existing versions: {versions}"
            )

    return str(module_version)


def _check_use_local_syntax_valid(use_local_list) -> bool:
    use_local_except_for = True if use_local_list[0].startswith("!") else False
    if use_local_except_for:
        for module_key in use_local_list:
            if not module_key.startswith("!"):
                return False
    else:
        for module_key in use_local_list:
            if module_key.startswith("!"):
                return False
    return True



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Compliant variant of tqdm.
"""


from tqdm import tqdm
from tqdm.utils import disp_len, _unicode
from typing import Iterable, Optional, Union
from io import TextIOWrapper, StringIO
import sys

from shrike.compliant_logging.constants import DataCategory
from shrike.compliant_logging.logging import is_eyesoff, get_prefix

# https://stackoverflow.com/a/65655249
# Number = Union[float, int]

# https://stackoverflow.com/a/50928627
# int --> float


class compliant_tqdm(tqdm):
    """
    Compliant `tqdm` wrapper. In order to prevent scrubbing of the `tqdm`
    progress bar in the eyes-off setting, this wrapper behaves differently when
    `category=DataCategory.PUBLIC` and the environment is eyes-off. In that
    case, the progress bar uses the `"SystemLog:"` prefix (used for system
    metadata that are not scrubbed), and it avoids the use of the carriage
    return (i.e., `"\r"`) prefix when displaying the progress bar, resulting in
    it not updating in place, but rather printing an updated bar in a new line.
    If `category=DataCategory.PRIVATE` (or the environment is eyes-on) then this
    wrapper behaves identically to `tqdm`.

    The simplest way to change the bar prefix is to use this argument when
    constructing it: `bar_format="SystemLog: {l_bar}{bar}{r_bar}"`.
    """

    def __init__(
        self,
        iterable: Optional[Iterable] = None,
        desc: Optional[str] = None,
        total: Optional[float] = None,
        leave: bool = True,
        file: Union[StringIO, TextIOWrapper, None] = None,
        ncols: Optional[int] = None,
        mininterval: Optional[float] = 0.1,
        maxinterval: Optional[float] = 10.0,
        miniters: Optional[float] = None,
        ascii: Union[bool, str, None] = None,
        disable: bool = False,
        unit: str = "it",
        unit_scale: Union[bool, float] = False,
        dynamic_ncols: bool = False,
        smoothing: float = 0.3,
        bar_format: Optional[str] = None,
        initial: float = 0,
        position: Optional[int] = None,
        postfix: Optional[dict] = None,
        unit_divisor: float = 1000,
        write_bytes: Optional[bool] = None,
        lock_args=None,
        nrows: Optional[int] = None,
        colour: Optional[str] = None,
        delay: Optional[float] = 0,
        gui: bool = False,
        category: DataCategory = DataCategory.PRIVATE,
        **kwargs,
    ):
        """
        Parameters:

        - `iterable: iterable`, optional. Iterable to decorate with a
          progressbar. Leave blank to manually manage the updates.
        - `desc: str`, optional. Prefix for the progressbar.
        - `total: int or float`, optional. The number of expected iterations. If
          unspecified, `len(iterable)` is used if possible. If `float("inf")` or
          as a last resort, only basic progress statistics are displayed. (no
          ETA, no progressbar). If `gui` is `True` and this parameter needs
          subsequent updating, specify an initial arbitrary large positive
          number, e.g. `9e9`.
        - `leave: bool`, optional. If `[default: True]`, keeps all traces of the
          progressbar upon termination of iteration. If `None`, will leave only
          if `position` is `0`.
        - `file: io.TextIOWrapper or io.StringIO`, optional. Specifies where to
          output the progress messages (default: `sys.stderr`). Uses
          `file.write(str)` and `file.flush()` methods. For encoding, see
          `write_bytes`.
        - `ncols: int`, optional. The width of the entire output message. If
          specified, dynamically resizes the progressbar to stay within this
          bound. If unspecified, attempts to use environment width. The fallback
          is a meter width of 10 and no limit for the counter and statistics. If
          0, will not print any meter (only stats).
        - `mininterval: float`, optional. Minimum progress display update
          interval [default: 0.1] seconds.
        - `maxinterval: float`, optional. Maximum progress display update
          interval [default: 10] seconds. Automatically adjusts `miniters` to
          correspond to `mininterval` after long display update lag. Only works
          if `dynamic_miniters` or monitor thread is enabled.
        - `miniters: int or float`, optional. Minimum progress display update
          interval, in iterations. If 0 and `dynamic_miniters`, will
          automatically adjust to equal `mininterval` (more CPU efficient, good
          for tight loops). If > 0, will skip display of specified number of
          iterations. Tweak this and `mininterval` to get very efficient loops.
          If your progress is erratic with both fast and slow iterations
          (network, skipping items, etc) you should set `miniters=1`.
        - `ascii: bool or str`, optional. If unspecified or False, use unicode
          (smooth blocks) to fill the meter. The fallback is to use ASCII
          characters `"123456789#"`.
        - `disable: bool`, optional. Whether to disable the entire progressbar
          wrapper [default: False]. If set to None, disable on non-TTY.
        - `unit: str`, optional. String that will be used to define the unit of
          each iteration [default: it].
        - `unit_scale: bool or int or float`, optional. If 1 or True, the number
          of iterations will be reduced/scaled automatically and a metric prefix
          following the International System of Units standard will be added
          (kilo, mega, etc.) [default: False]. If any other non-zero number,
          will scale `total` and `n`.
        - `dynamic_ncols: bool`, optional. If set, constantly alters `ncols` and
          `nrows` to the environment (allowing for window resizes) [default:
          False].
        - `smoothing`: float`, optional. Exponential moving average smoothing
          factor for speed estimates (ignored in GUI mode). Ranges from 0
          (average speed) to 1 (current/instantaneous speed) [default: 0.3].
        - `bar_format: str`, optional. Specify a custom bar string formatting.
          May impact performance. [default: '{l_bar}{bar}{r_bar}'], where
          `l_bar='{desc}: {percentage:3.0f}%|'` and
          `r_bar='| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, '`
          `'{rate_fmt}{postfix}]'`.
          Possible vars: `l_bar`, `bar`, `r_bar`, `n`, `n_fmt`, `total`,
          `total_fmt`, `percentage`, `elapsed`, `elapsed_s`, `ncols`, `nrows`,
          `desc`, `unit`, `rate`, `rate_fmt`, `rate_noinv`, `rate_noinv_fmt`,
          `rate_inv`, `rate_inv_fmt`, `postfix`, `unit_divisor`, `remaining`,
          `remaining_s`, eta. Note that a trailing ": " is automatically removed
          after `{desc}` if the latter is empty.
        - `initial: int or float`, optional. The initial counter value. Useful
          when restarting a progress bar [default: 0]. If using float, consider
          specifying `{n:.3f}` or similar in `bar_format`, or specifying
          `unit_scale`.
        - `position: int`, optional. Specify the line offset to print this bar
          (starting from 0). Automatic if unspecified. Useful to manage multiple
          bars at once (eg, from threads).
        - `postfix: dict or *`, optional. Specify additional stats to display at
          the end of the bar. Calls `set_postfix(**postfix)` if possible (dict).
        - `unit_divisor: float`, optional. [default: 1000], ignored unless
          `unit_scale` is True.
        - `write_bytes: bool`, optional. If (default: None) and `file` is
          unspecified, bytes will be written in Python 2. If `True` will also
          write bytes. In all other cases will default to unicode.
        - `lock_args: tuple`, optional. Passed to `refresh` for intermediate
          output (initialisation, iterating, and updating).
        - `nrows: int`, optional. The screen height. If specified, hides nested
          bars outside this bound. If unspecified, attempts to use environment
          height. The fallback is 20.
        - `colour: str`, optional. Bar colour (e.g. 'green', `'#00ff00'`).
        - `delay: float`, optional. Don't display until [default: 0] seconds
          have elapsed.
        - gui: bool`, optional. WARNING: internal parameter - do not use. Use
          `tqdm.gui.tqdm(...)` instead. If set, will attempt to use matplotlib
          animations for a graphical output [default: False].
        """

        if category == DataCategory.PUBLIC and is_eyesoff():
            if not bar_format:
                bar_format = "{l_bar}{bar}{r_bar}"
            bar_format = f"{get_prefix()}{bar_format}"

        super().__init__(
            iterable=iterable,
            desc=desc,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )

    @staticmethod
    def status_printer(file):
        fp = file
        fp_flush = getattr(fp, "flush", lambda: None)  # pragma: no cover
        if fp in (sys.stderr, sys.stdout):
            sys.stderr.flush()
            sys.stdout.flush()

        def fp_write(value: str) -> None:
            fp.write(_unicode(value))
            fp_flush()

        last_len = [0]

        prefix = get_prefix()

        def print_status(status: str) -> None:
            fp_write(f"{prefix} ")
            if prefix and status.startswith(prefix):
                fp_write(f"{status}\n")
            else:
                len_s = disp_len(status)
                fp_write("\r" + status + (" " * max(last_len[0] - len_s, 0)))
                last_len[0] = len_s

        return print_status



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import argparse
import torch
import os


def get_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--model_output_dir", help="model output directory", required=True
    )

    for i in range(1, 11):
        input_name = "--model_input_dir_" + str(i)
        parser.add_argument(
            input_name, type=str, required=i <= 2
        )  # only the first two inputs are required

    parser.add_argument("--num_models", type=int, required=True)
    parser.add_argument("--weights", type=str, required=False)
    parser.add_argument(
        "--model_name", type=str, required=False, default="pytorch_model.bin"
    )

    args, unknown = parser.parse_known_args()
    return args


def main(args):
    if args.weights:
        weights = [float(i) for i in args.weights.split(",")]
        try:
            assert sum(weights) == 1.0
        except AssertionError as e:
            print("SystemLog: the sum of weights is unequal to 1.")
            raise
        try:
            assert len(weights) == args.num_models
        except AssertionError as e:
            print(
                "SystemLog: the number of weights does not match with the number of input models."
            )
            raise
    else:
        weights = 1 / args.num_models

    models = []
    for i in range(args.num_models):
        input_name = "model_input_dir_" + str(i + 1)
        input_path = vars(args)[input_name]
        if os.path.isdir(input_path):
            input_path = os.path.join(input_path, args.model_name)
        models.append(torch.load(input_path, map_location=torch.device("cpu")))

    agg_model = models[0].copy()
    for k, v in agg_model.items():
        if ".embeddings." in k or "encoder.layer.0." in k or "encoder.layer.1." in k:
            # for freezed layers, not change it
            continue
        agg_model[k] = agg_model[k] * 0
        for i in range(args.num_models):
            agg_model[k] = agg_model[k] + models[i][k] * weights[i]

    output_path_to_agg_model = os.path.join(args.model_output_dir, args.model_name)
    torch.save(agg_model, output_path_to_agg_model)
    del agg_model


if __name__ == "__main__":
    print("SystemLog: start parsing args")
    try:
        args = get_args()
    except:
        print("SystemLog: got error when parsing args!")
    print("SystemLog: done get args, parsed args are: {}".format(args))
    try:
        main(args)
    except Exception as e:
        print("SystemLog: got error when running main(args), error message: " + str(e))
